import('lang.Object')

---@class model.Component : lang.Object
---@field protected __componentType string
---@field private _root LuaGuiElement
Component = extends(Object)
Component.__componentType = nil

---[static]
---@public
---@param parent LuaGuiElement
---@return self
function Component:create(parent)
    error("Not implemented")
end

---@public
---@param guiEvent model.GuiEvent
function Component:handle(guiEvent)
    local child = self:findChild(guiEvent:element())
    if child then
        child:handle(guiEvent)
    end
end

---@protected
---@param element LuaGuiElement
function Component:findChild(element)
    for _, child in ipairs(self:childrenClasses()) do
        if child:contains(element) then
            return child:findUp(element)
        end
    end
end

---@public
function Component:update()
    for _, child in ipairs(self:children()) do
        child:update()
    end
end

function Component:lock()
    for _, child in ipairs(self:children()) do
        child:lock()
    end
end

function Component:unlock()
    for _, child in ipairs(self:children()) do
        child:unlock()
    end
end

---@private
---@return model.Component[]
function Component:children()
    local children = {}
    for _, class in ipairs(self:childrenClasses()) do
        for _, child in ipairs(class:findDown(self:root())) do
            table.insert(children, child)
        end
    end
    return children
end

---@return model.Component[]
function Component:childrenClasses()
    return {}
end

---[static]
---Sets element.tags.componentType = self.__type to recognize it while searching
---@protected
---@param parent LuaGuiElement
---@param addParameters LuaGuiElement.add_parameters
---@return LuaGuiElement
function Component:createRootElement(parent, addParameters)
    addParameters.tags = { componentType = self.__componentType }
    local element = parent.add(addParameters)
    return element
end

---@public
---@param element LuaGuiElement
---@return boolean
function Component:contains(element)
    return self:findUp(element) ~= nil
end

---[static]
---Recognizes itself by element.tags.componentType set in Component:createRootElement
---@public
---@param element LuaGuiElement
---@return self
function Component:findUp(element)
    if self:is(element) then
        return self:new(element)
    elseif element.parent then
        return self:findUp(element.parent)
    else
        return nil
    end
end

function Component:findDownFirst(element)
    return self:findDown(element)[1]
end

---@public
---@param element LuaGuiElement
---@return self[]
function Component:findDown(element)
    if self:is(element) then
        return { self:new(element) }
    elseif element.children then
        local founded = {}
        for _, child in pairs(element.children) do
            local foundedInChildren = self:findDown(child)
            for _, component in ipairs(foundedInChildren) do
                table.insert(founded, component)
            end
        end
        return founded
    else
        return {}
    end
end

---[static]
---@public
---@param element LuaGuiElement
---@return boolean
function Component:is(element)
    assert(self.__componentType ~= nil, "Component must have defined __componentType")
    return element.tags and element.tags.componentType == self.__componentType
end

---[static]
---@protected
---@param element LuaGuiElement
---@return self
function Component:new(element)
    local this = Component:super(Object:new())
    this._root = Object.requireNonNil(element)
    return this
end

---@protected
---@param replacement model.Component
function Component:replaceWith(replacement)
    self:parent().swap_children(self:index(), replacement:index())
    self:delete()
end

---@protected
---@param other model.Component
function Component:swapWith(other)
    self:parent().swap_children(self:index(), other:index())
end

---@public
---@return boolean
function Component:isValid()
    return self:root().valid
end

---@public
function Component:show()
    self:root().visible = true
end

---@public
function Component:hide()
    self:root().visible = false
end

---@public
function Component:delete()
    self:root().destroy()
end

---@protected
---@return number
function Component:index()
    return self:root().get_index_in_parent()
end

---@public
---@return self[]
function Component:siblings()
    return self:findDown(self:parent())
end

---@protected
---@return LuaGuiElement
function Component:parent()
    return self:root().parent
end

---@protected
---@return LuaGuiElement
function Component:root()
    return self._root
end
